#!/usr/bin/env python3

"""
dycall.exports
~~~~~~~~~~~~~~

Contains `ExportsFrame` and `ExportsTreeView`.
"""

from __future__ import annotations

import logging
import pathlib
from typing import TYPE_CHECKING

import ttkbootstrap as tk
from ttkbootstrap import ttk
from ttkbootstrap.dialogs import Messagebox
from ttkbootstrap.localization import MessageCatalog as MsgCat
from ttkbootstrap.tableview import Tableview

from dycall.types import Export, PEExport
from dycall.util import StaticThemedTooltip, get_img

log = logging.getLogger(__name__)


class ExportsFrame(ttk.Labelframe):
    """Contains **Exports** combobox and a button for `ExportsTreeView`.

    Use command line argument `--exp` to select an export from the library on
    launch. Combobox validates export name.

    TODO: Combobox works like google search (auto-suggest, recents etc.)
    """

    def __init__(
        self,
        root: tk.Window,
        selected_export: tk.StringVar,
        sort_order: tk.StringVar,
        output: tk.StringVar,
        status: tk.StringVar,
        is_loaded: tk.BooleanVar,
        is_native: tk.BooleanVar,
        is_reinitialised: tk.BooleanVar,
        lib_path: tk.StringVar,
        exports: list[Export],
    ):
        log.debug("Initalising")

        super().__init__(text=MsgCat.translate("Exports"))
        self.__root = root
        self.__selected_export = selected_export
        self.__sort_order = sort_order
        self.__output = output
        self.__status = status
        self.__is_loaded = is_loaded
        self.__is_native = is_native
        self.__is_reinitialised = is_reinitialised
        self.__lib_path = lib_path
        self.__exports = exports
        self.__export_names: list[str] = []

        self.cb = cb = ttk.Combobox(
            self,
            state="disabled",
            textvariable=selected_export,
            validate="focusout",
            validatecommand=(self.register(self.cb_validate), "%P"),
        )
        # ! cb.bind("<Return>", lambda *_: self.cb_validate)  # Doesn't work
        cb.bind("<<ComboboxSelected>>", self.cb_selected)

        self.__list_png = get_img("list.png")
        self.lb = lb = ttk.Label(self, image=self.__list_png)
        lb.bind(
            "<Enter>",
            lambda *_: StaticThemedTooltip(lb, MsgCat.translate("List of exports")),
        )
        lb.bind("<ButtonRelease-1>", lambda *_: status.set("Load a library first!"))

        lb.pack(padx=(0, 5), pady=5, side="right")
        cb.pack(fill="x", padx=5, pady=5)

        self.bind_all("<<PopulateExports>>", lambda *_: self.set_cb_values())
        self.bind_all(
            "<<ToggleExportsFrame>>", lambda event: self.set_state(event.state == 1)
        )
        self.bind_all("<<SortExports>>", lambda *_: self.sort())
        log.debug("Initialised")

    def cb_selected(self, *_):
        """Callback to handle clicks on **Exports** combobox.

        Resets **Output** and activates/deactivates `FunctionFrame`.
        """
        log.debug("%s selected", self.__selected_export.get())
        self.__output.set("")
        if self.__is_native.get():
            self.__root.event_generate("<<ToggleFunctionFrame>>", state=1)
        else:
            self.__root.event_generate("<<ToggleFunctionFrame>>", state=0)

    def cb_validate(self, *_) -> bool:
        """Callback to handle keyboard events on **Exports** combobox.

        Activates `FunctionFrame` when the text in the combobox
        is a valid export name. Deactivates it otherwise.
        """
        log.debug("Validating Exports combobox")
        try:
            # Don't validate if combobox dropdown arrow was pressed
            self.cb.state()[1] == "pressed"
        except IndexError:
            exp = self.cb.get()
            if exp:
                if exp in self.__export_names:
                    self.cb_selected()
                    return True
                self.__root.event_generate("<<ToggleFunctionFrame>>", state=1)
                return False
        return True

    def set_state(self, activate: bool = True):
        """Activates/deactivates **Exports** combobox.

        Args:
            activate (bool, optional): Activated when True, deactivated when
                False. Defaults to True.
        """
        log.debug("Called with activate=%s", activate)
        state = "normal" if activate else "disabled"
        self.cb.configure(state=state)

    def set_cb_values(self):
        """Demangles and sets the export names to the **Exports** combobox."""
        exports = self.__exports
        if not self.__is_reinitialised.get() or self.__is_loaded.get():
            num_exports = len(exports)
            log.info("Found %d exports", num_exports)
            self.__status.set(f"{num_exports} exports found")
            failed = []
            for exp in exports:
                if isinstance(exp, PEExport):
                    if hasattr(exp, "exc"):
                        failed.append(exp.name)
            if failed:
                Messagebox.show_warning(
                    f"These export names couldn't be demangled: {failed}",
                    "Demangle Errors",
                    parent=self.__root,
                )
        self.__export_names = names = list(e.demangled_name for e in exports)
        self.set_state()
        self.cb.configure(values=names)
        selected_export = self.__selected_export.get()
        if selected_export:
            if selected_export not in names:
                err = "%s not found in export names"
                log.error(err, selected_export)
                Messagebox.show_error(
                    err % selected_export, "Export not found", parent=self.__root
                )
                self.cb.set("")
            else:
                # Activate function frame when export name is passed from command line
                self.cb_selected()
        self.lb.configure(cursor="hand2")
        self.lb.bind(
            "<ButtonRelease-1>",
            lambda *_: ExportsTreeView(
                self.__exports, pathlib.Path(self.__lib_path.get()).name
            ),
            add=False,
        )

    def sort(self, *_):
        """Sorts the list of export names and repopulates the combobox."""
        if self.__is_loaded.get():
            sorter = self.__sort_order.get()
            log.debug("Sorting w.r.t. %s", sorter)
            names = self.__export_names
            if sorter == "Name (ascending)":
                names.sort()
            elif sorter == "Name (descending)":
                names.sort(reverse=True)
            self.cb.configure(values=names)
        self.__status.set("Sort order changed")


class ExportsTreeView(tk.Toplevel):
    """Displays detailed information about all the exports of a library.

    Following information is displayed:
    - Address
    - Name
    - Demangled name (whenever available)
    - Ordinal (Windows only)
    """

    def __init__(self, exports: list[Export], lib_name: str):
        log.debug("Initialising")
        super().__init__(
            title=f"{MsgCat.translate('Exports')} - {lib_name}", size=(400, 500)
        )
        self.__old_height = 0
        self.withdraw()
        coldata = [
            "Address",
            "Name",
            {"text": "Demangled", "stretch": True},
        ]
        is_pe = isinstance(exports[0], PEExport)
        if is_pe:
            coldata.insert(0, "Ordinal")
        self.__tv = tv = Tableview(
            self,
            searchable=True,
            autofit=True,
            coldata=coldata,
            paginated=True,
            pagesize=25,
        )
        tv.pack(fill="both", expand=True)
        for e in exports:
            values = [e.address, e.name, e.demangled_name]  # type: ignore
            if is_pe:
                if TYPE_CHECKING:
                    assert isinstance(e, PEExport)  # nosec
                values.insert(0, e.ordinal)
            tv.insert_row(values=values)
        tv.load_table_data()
        self.bind(
            "<F11>",
            lambda *_: self.attributes(
                "-fullscreen", not self.attributes("-fullscreen")
            ),
        )
        self.bind("<Configure>", self.resize)
        self.deiconify()
        self.focus_set()
        log.debug("Initialised")

    def resize(self, event: tk.tk.Event):
        """Change the treeview's `pagesize` whenever this window is resized.

        I came up with this because there is no way to show a vertical
        scrollbar in/for the treeview.
        """
        new_height = event.height
        if event.widget.widgetName == "toplevel" and new_height != self.__old_height:
            # ! This is an expensive call, avoid it whenever possible
            self.__tv.pagesize = int(new_height) / 20
            self.__old_height = new_height



#!/usr/bin/env python3

"""
dycall.output
~~~~~~~~~~~~~

Contains `OutputFrame`.
"""

import logging

import ttkbootstrap as tk
from ttkbootstrap import ttk
from ttkbootstrap.localization import MessageCatalog as MsgCat

from dycall.util import CopyButton

log = logging.getLogger(__name__)


class OutputFrame(ttk.Labelframe):
    """Shows the value returned/exception caused by calling the exported function.

    Contains a readonly `Entry` to display text and a copy button alongside.
    """

    def __init__(self, _: tk.Window, output: tk.StringVar, exc_type: tk.StringVar):
        log.debug("Initialising")
        title = MsgCat.translate("Output")
        super().__init__(text=title)
        self.bind_all("<<OutputSuccess>>", lambda *_: self.configure(text=title))
        self.bind_all(
            "<<OutputException>>", lambda _: self.configure(text=exc_type.get())
        )
        self.oe = oe = ttk.Entry(
            self,
            font="TkFixedFont",
            state="readonly",
            textvariable=output,
        )
        self.oc = oc = CopyButton(self, output, state="disabled")
        oc.pack(side="right", padx=(0, 5), pady=5)
        oe.pack(fill="x", padx=5, pady=5)
        log.debug("Initialised")



#!/usr/bin/env python3

"""
dycall.runner
~~~~~~~~~~~~~

Contains `Runner`.
"""

from __future__ import annotations

import ctypes
import logging
import platform
import queue
import threading
from typing import Union

import ttkbootstrap as tk

from dycall.types import CallConvention, Marshaller, ParameterType, RunResult

log = logging.getLogger(__name__)


class Runner(threading.Thread):
    """Executes an exported function in a separate thread.

    Used in `FunctionFrame`. Exceptions and results are pushed into a queue.
    The queues are then checked regularly until they are not empty in the UI
    thread. This ensures that the UI doesn't get blocked.
    """

    def __init__(
        self,
        exc: queue.Queue,
        que: queue.Queue,
        args: list[list[str]],
        call_conv: str,
        returns: str,
        lib_path: str,
        name_or_ord: str,
        get_last_error: tk.IntVar,
        show_get_last_error: bool,
        errno: tk.IntVar,
        show_errno: bool,
    ) -> None:
        log.debug(
            "Called with args=%s, "
            "call_conv=%s, "
            "returns=%s, "
            "lib_path=%s, "
            "name_or_ord=%s, "
            "get_last_error=%d, "
            "show_get_last_error=%s, "
            "errno=%d, "
            "show_errno=%s",
            args,
            call_conv,
            returns,
            lib_path,
            name_or_ord,
            get_last_error.get(),
            show_get_last_error,
            errno.get(),
            show_errno,
        )
        self.__exc = exc
        self.__queue = que
        self.__get_last_error = get_last_error
        self.__show_get_last_error = show_get_last_error
        self.__errno = errno
        self.__show_errno = show_errno
        self.__is_windows = platform.system() == "Windows"
        self.__call_conv = CallConvention(call_conv)
        self.__restype = ParameterType(returns).ctype
        if self.__call_conv == CallConvention.StdCall:
            self.__handle = ctypes.WinDLL(
                lib_path,
                use_last_error=show_get_last_error,
                use_errno=show_errno,
            )
            self.__functype = ctypes.WINFUNCTYPE
        else:
            if self.__is_windows:
                self.__handle = ctypes.CDLL(  # type: ignore
                    lib_path,
                    use_errno=show_errno,
                    use_last_error=show_get_last_error,
                )
            else:
                self.__handle = ctypes.CDLL(  # type: ignore
                    lib_path, use_errno=show_errno
                )
            self.__functype = ctypes.CFUNCTYPE
        if name_or_ord.startswith("@"):
            self.__name_or_ord = int(name_or_ord[1:])  # type: Union[str, int]
        else:
            self.__name_or_ord = name_or_ord
        self.__argtypes = []
        self.__argvalues = []
        for arg in args:
            type_, value = arg
            argtype = ParameterType(type_).ctype
            argvalue = Marshaller.str2ctype(argtype, value)
            self.__argtypes.append(argtype)
            self.__argvalues.append(argvalue)
        super().__init__()

    def run(self):
        """Calculates the function prototype and operates with the queues."""
        try:
            if self.__argtypes:
                prototype = self.__functype(self.__restype, *self.__argtypes)
            else:
                prototype = self.__functype(self.__restype)
            ptr = prototype((self.__name_or_ord, self.__handle))
            result = ptr(*self.__argvalues)
            run_result = RunResult(result, self.__argvalues)
        except Exception as e:  # pylint: disable=broad-except
            self.__exc.put(e)
        else:
            self.__queue.put(run_result)
        # This is thread safe, see https://stackoverflow.com/a/25352087
        if self.__show_get_last_error and self.__is_windows:
            # ! ctypes.get_last_error() doesn't work
            gle = int(ctypes.windll.kernel32.GetLastError())
            log.debug("GetLastError - %d", gle)
            self.__get_last_error.set(gle)
        if self.__show_errno:
            errno = ctypes.get_errno()
            log.debug("errno - %d", errno)
            self.__errno.set(errno)



#!/usr/bin/env python3

"""Pytest fixtures"""

import appdirs
import pytest

from dycall.app import App


@pytest.fixture
def create_app(monkeypatch, tmp_path):
    """Instantiates DyCall with a monkeypatched config dir."""

    def mock_config_dir(*_):
        return tmp_path

    monkeypatch.setattr(appdirs, "user_config_dir", mock_config_dir)
    return App()



#!/usr/bin/env python3

"""Adds padding to a square image an resizes it back to original resolution.

Use this for creating top menu icons. Resulting image is overwritten.
Ref: https://note.nkmk.me/en/python-pillow-add-margin-expand-canvas/
"""

import argparse
import pathlib

from PIL import Image


def main(img: pathlib.Path, padding: int):  # noqa
    with Image.open(img) as image:
        width, height = image.size
        resized = image.resize((width - (padding * 2), height - (padding * 2)))
        padded_image = Image.new(image.mode, (width, height))
        padded_image.paste(resized, (padding, padding))
    padded_image.save(img)


if __name__ == "__main__":
    ap = argparse.ArgumentParser(description=__doc__)
    ap.add_argument("img", type=pathlib.Path)
    ap.add_argument("padding", type=int)
    args = vars(ap.parse_args())
    main(**args)



#!/usr/bin/env python3

"""
dycall.app
~~~~~~~~~~

Contains `App`.
"""

from __future__ import annotations

import collections
import logging
import os
import pathlib
import platform

try:
    from typing import Final  # type: ignore
except ImportError:
    from typing_extensions import Final  # type: ignore

import appdirs
import darkdetect
import easysettings
import ttkbootstrap as tk
from ttkbootstrap.dialogs import Messagebox
from ttkbootstrap.icons import Icon
from ttkbootstrap.localization import MessageCatalog as MsgCat

import dycall.util
from dycall.exports import ExportsFrame
from dycall.function import FunctionFrame
from dycall.output import OutputFrame
from dycall.picker import PickerFrame
from dycall.status_bar import StatusBarFrame
from dycall.top_menu import TopMenu
from dycall.types import CallConvention, Export, SortOrder
from dycall.util import DARK_THEME, LIGHT_THEME, get_img_path

log = logging.getLogger(__name__)

# https://stackoverflow.com/a/3430395
dirpath = pathlib.Path(__file__).parent.resolve()


class App(tk.Window):
    """Welcome to DyCall!

    Handles the creation of all subframes, virtual events and control variables.
    Command line arguments are parsed if present and application configuration
    is managed.

    Most analogous to the "Controller" part of the MVC design pattern.

    Events:
        - <<LanguageChanged>>
        - <<OutputSuccess>>
        - <<OutputException>>
        - <<PopulateExports>>
        - <<SortExports>>
        - <<ThemeChanged2>>: https://github.com/ragardner/tksheet/issues/111
        - <<ToggleErrno>>
        - <<ToggleExportsFrame>>
        - <<ToggleFunctionFrame>>
        - <<ToggleGetLastError>>
        - <<UpdateRecents>>

    Control variables:
        about_opened (tk.BooleanVar): Whether `dycall.about.AboutWindow` is open.
        is_loaded (tk.BooleanVar): Whether a library has been selected.
        is_native (tk.BooleanVar): Whether loaded library is native.
        is_running (tk.BooleanVar): Set to True when a function is executing
            and False again after it completes execution. Defaults to False.
    """

    def __init__(
        self,
        conv: str = CallConvention.Cdecl.value,
        exp: str = "",
        lib: str = "",
        ret: str = "",
        rows: int = 0,
        lang: str = "",
        out_mode: bool = False,
        hide_gle: bool = False,
        hide_errno: bool = False,
        no_images: bool = False,
    ) -> None:
        """DyCall entry point.

        Certain arguments can be passed from the command line directly. These
        arguments will be used to initalise the interface. This helps achieve
        some automation and it also saves my time while testing. The default
        arguments aren't the default values used by DyCall interface.

        Args:
            conv (str, optional): Calling convention. Defaults to "cdecl".
            exp (str, optional): Name of exported function. Defaults to "".
            lib (str, optional): Path/name of the library. Defaults to "".
            ret (str, optional): Return type. See `ParameterType`. Defaults to "".
            rows (int, optional): Number of empty rows to add to arguments table.
                Defaults to 0.
            lang (str, optional): The language used for displaying UI. Defaults to "".
            out_mode (bool, optional): Whether **OUT Mode** should be enabled.
                Defaults to False.
            hide_gle (bool, optional): Hides and doesn't retrieve GetLastError
                value shown in status bar. (Windows only)
            hide_errno (bool, optional): Hides and doesn't retrieve errno value
                shown in status bar.
            no_images (bool, optional): Opens the DyCall GUI without loading
                any images
        """  # noqa: D403
        log.debug("Initialising")

        # Set icon
        log.debug("Setting app icon")
        if platform.system() != "Windows":
            super().__init__(iconphoto=get_img_path("dycall.png"))
        else:
            super().__init__(iconphoto=None)
            ico = dirpath / "img/dycall.ico"
            self.iconbitmap(ico, ico)
        self.withdraw()

        # Load config
        log.debug("Loading config")
        configdir = appdirs.user_config_dir("DyCall", "demberto")
        if not os.path.exists(configdir):
            log.info("Config dir doesn't existing, creating it...")
            os.makedirs(configdir)
        configpath = os.path.join(configdir, "settings.json")
        self.__config = config = easysettings.load_json_settings(
            configpath,
            default={
                "theme": "System",
                "geometry": self.centre_position,
                "out_mode": False,
                "locale": "en",
                "recents": [],
                "show_get_last_error": True,
                "show_errno": True,
            },
        )

        # No need to save a preference in the config when
        # DyCall is passed with it from the command line.
        self.__dont_save_locale = False
        self.__dont_save_out_mode = False
        self.__dont_save_show_get_last_error = False
        self.__dont_save_show_errno = False

        if lang:
            locale_to_use = lang
            self.__dont_save_locale = True
        else:
            locale_to_use = config["locale"]
        if out_mode:
            out_mode_or_not = True
            self.__dont_save_out_mode = True
        else:
            out_mode_or_not = config["out_mode"]
        if hide_gle:
            show_get_last_error = False
            self.__dont_save_show_get_last_error = True
        else:
            show_get_last_error = config["show_get_last_error"]
        if hide_errno:
            show_errno = False
            self.__dont_save_show_errno = True
        else:
            show_errno = config["show_errno"]

        # Load translations
        # ! BUG: Tkinter doesn't understand Windows paths
        msgs_path = os.path.join(dirpath, "msgs").replace("\\", "/")

        # * The order and place where these 2 are called is very important
        log.debug("Loading message catalogs from %s", msgs_path)
        found = MsgCat.load(msgs_path)
        log.debug("Found %d translation files", found)
        MsgCat.locale(locale_to_use)

        # Control variables
        self.__cur_theme: Final = tk.StringVar(value=config["theme"])
        self.__is_native: Final = tk.BooleanVar()
        self.__is_running: Final = tk.BooleanVar(value=False)
        self.__is_loaded: Final = tk.BooleanVar(value=False)
        self.__is_reinitialised: Final = tk.BooleanVar(value=False)
        self.__use_out_mode: Final = tk.BooleanVar(value=out_mode_or_not)
        self.__locale: Final = tk.StringVar(value=locale_to_use)
        self.__sort_order: Final = tk.StringVar(value=SortOrder.NameAscending.value)
        self.__lib_path: Final = tk.StringVar(value=lib)
        self.__selected_export: Final = tk.StringVar(value=exp)
        self.__call_convention: Final = tk.StringVar(value=conv)
        self.__return_type: Final = tk.StringVar(value=ret)
        self.__output_text: Final = tk.StringVar()
        self.__status_text: Final = tk.StringVar(
            value=MsgCat.translate("Choose a library")
        )
        self.__exc_type: Final = tk.StringVar()
        self.__get_last_error: Final = tk.IntVar()
        self.__show_get_last_error: Final = tk.BooleanVar(value=show_get_last_error)
        self.__errno: Final = tk.IntVar()
        self.__show_errno: Final = tk.BooleanVar(value=show_errno)
        self.__about_opened: Final = tk.BooleanVar(value=False)

        # Events
        # ! Always use `self.__parent.event_generate`
        self.bind_all("<<LanguageChanged>>", lambda *_: self.refresh())
        self.bind_all(
            "<<SetWindowTitle>>",
            lambda *_: self.title(
                f"{self.__default_title} - {pathlib.Path(self.__lib_path.get()).name}"
            ),
        )
        self.bind("<<ThemeChanged2>>", lambda *_: self.set_theme())
        self.bind(
            "<F11>",
            lambda *_: self.attributes(
                "-fullscreen", not self.attributes("-fullscreen")
            ),
        )

        # Control variable traces
        self.__show_get_last_error.trace_add(
            "write", lambda *_: self.event_generate("<<ToggleGetLastError>>")
        )
        self.__show_errno.trace_add(
            "write", lambda *_: self.event_generate("<<ToggleErrno>>")
        )

        # Misc
        self.__no_images = no_images
        if no_images:
            log.debug(
                "Images will not be shown. Run DyCall without "
                "the '--no-images' option to see them."
            )
            # https://stackoverflow.com/a/3536638
            dycall.util.SHOW_IMAGES = False
        self.__arch: Final = platform.architecture()[0]
        self.__rows_to_add: Final = rows
        self.__exports: list[Export] = []
        self.__recents: Final[collections.deque] = collections.deque(
            config["recents"], maxlen=10
        )
        self.title(self.__default_title)
        self.minsize(width=450, height=600)
        self.geometry(config["geometry"])
        self.option_add("*tearOff", False)  # Modern menus
        self.set_theme()
        self.init_widgets()
        self.deiconify()
        log.debug("App initialised")

    @property
    def __default_title(self) -> str:
        title = f"DyCall ({self.__arch})"
        if self.__no_images:
            title += " [IMAGELESS MODE]"
        return title

    def init_widgets(self):
        """Sub-widgets are created and packed here.

        Separated from `__init__` because `refresh` requires this method too.
        These widgets are purposely public to allow subframes to access each
        other's widgets through their `parent` attribute when necessary. The
        order in which they get constructed is essential, it must be ensured
        that no widget constructed first depends internally on a widget
        constructed later here.

        Call this AFTER `set_theme` else **Arguments** table will mess up.
        """
        self.output = of = OutputFrame(
            self,
            self.__output_text,
            self.__exc_type,
        )
        self.status_bar = sf = StatusBarFrame(
            self,
            self.__status_text,
            self.__get_last_error,
            self.__show_get_last_error,
            self.__errno,
            self.__show_errno,
        )
        self.function = ff = FunctionFrame(
            self,
            self.__call_convention,
            self.__return_type,
            self.__lib_path,
            self.__selected_export,
            self.__output_text,
            self.__status_text,
            self.__use_out_mode,
            self.__is_running,
            self.__exc_type,
            self.__get_last_error,
            self.__show_get_last_error,
            self.__errno,
            self.__show_errno,
            self.__rows_to_add,
        )
        self.exports = ef = ExportsFrame(
            self,
            self.__selected_export,
            self.__sort_order,
            self.__output_text,
            self.__status_text,
            self.__is_loaded,
            self.__is_native,
            self.__is_reinitialised,
            self.__lib_path,
            self.__exports,
        )
        self.picker = pf = PickerFrame(
            self,
            self.__lib_path,
            self.__selected_export,
            self.__output_text,
            self.__status_text,
            self.__is_loaded,
            self.__is_native,
            self.__exports,
            self.__recents,
        )
        self.top_menu = tm = TopMenu(
            self,
            self.__use_out_mode,
            self.__locale,
            self.__sort_order,
            self.__show_get_last_error,
            self.__show_errno,
            self.__about_opened,
            self.__cur_theme,
            self.__recents,
        )

        pf.pack(fill="x", padx=5)
        ef.pack(fill="x", padx=5)
        ff.pack(fill="both", expand=True)  # Padding handled by the frame
        of.pack(fill="x", padx=5)
        sf.pack(fill="x")
        self["menu"] = tm

    def refresh(self):
        """Called when the interace language is changed to reflect the changes.

        TtkBootstrap doesn't let me destroy and reinitialise `App` so I found
        out this cool solution. Sets reinitalised flag so that re-demangling is
        prevented in `self.exports`.
        """
        self.__is_reinitialised.set(True)
        self.top_menu.destroy()
        self.picker.destroy()
        self.exports.destroy()
        self.function.destroy()
        self.output.destroy()
        self.status_bar.destroy()
        self.set_theme()
        self.init_widgets()

    def destroy(self):
        """Warns the user if he tries to close when an operation is running.
        Tries to save the app settings and proceeds to close the app.
        """
        # ! This does't work at all
        is_running = self.__is_running.get()
        log.debug("Called with is_running=%s", is_running)
        if self.__is_running.get():
            if (
                Messagebox.show_question(
                    "An operation is running, do you really want to quit?",
                    buttons=("No:primary", "Yes:danger"),
                )
                != "Yes"
            ):
                return

        config = self.__config
        config["theme"] = self.__cur_theme.get()
        config["geometry"] = self.geometry()
        config["recents"] = tuple(self.__recents)
        if not self.__dont_save_out_mode:
            config["out_mode"] = self.__use_out_mode.get()
        if not self.__dont_save_locale:
            config["locale"] = self.__locale.get()
        if not self.__dont_save_show_get_last_error:
            config["show_get_last_error"] = self.__show_get_last_error.get()
        if not self.__dont_save_show_errno:
            config["show_errno"] = self.__show_errno.get()
        try:
            config.save()
        except IOError as e:
            result = Messagebox.retrycancel(
                f"Failed to save config file due to {repr(e)}",
                "Error",
                parent=self,
                icon=Icon.error,
            )
            if result == "Retry":
                self.destroy()
        super().destroy()

    @property
    def centre_position(self):
        """Calculates the center position w.r.t. the screen."""
        w_height = self.winfo_height()
        w_width = self.winfo_width()
        s_height = self.winfo_screenheight()
        s_width = self.winfo_screenwidth()
        xpos = (s_width - w_width) // 2
        ypos = (s_height - w_height) // 2
        return f"+{xpos}+{ypos}"

    def set_theme(self):
        """Sets the theme used by DyCall."""
        log.debug("Setting theme")

        theme = self.__cur_theme.get()
        if theme == "System":
            theme = darkdetect.theme()  # pylint: disable=assignment-from-none

        if theme == "Light":
            self.style.theme_use(LIGHT_THEME)
        elif theme == "Dark":
            self.style.theme_use(DARK_THEME)

        log.debug("Theme '%s' set", theme)



#!/usr/bin/env python3

"""
dycall.demangler
~~~~~~~~~~~~~~~~

Contains `DemanglerWindow`.
"""

import logging

import ttkbootstrap as tk
from ttkbootstrap import ttk
from ttkbootstrap.dialogs import Messagebox
from ttkbootstrap.localization import MessageCatalog as MsgCat

from dycall.util import CopyButton, DemangleError, demangle

log = logging.getLogger(__name__)


class DemanglerWindow(tk.Toplevel):
    """A C++ symbol name demangler utility.

    Found under **Tools** -> **Demangler** in the top menu.
    Uses `dycall.util.demangle` to demangle the name.

    TODO: Syntax highlighted demangled name using Pygments.
    """

    def __init__(self, _: tk.Window):
        log.debug("Initialising")
        self.mangled_name = mangled_name = tk.StringVar()
        self.demangled_name = demangled_name = tk.StringVar()

        super().__init__(title="Demangler", toolwindow=True)
        self.withdraw()
        self.minsize(300, 100)
        self.resizable(True, False)
        self.geometry("500x100")

        self.columnconfigure(0)
        self.columnconfigure(1, weight=1)
        self.columnconfigure(2)
        self.rowconfigure(0, minsize=45, weight=1)
        self.rowconfigure(1, minsize=45, weight=1)

        self.ml = ml = ttk.Label(self, text=MsgCat.translate("Name"))
        self.me = me = ttk.Entry(self, textvariable=mangled_name)
        self.mb = mb = ttk.Button(
            self, text="Demangle", command=self.demangle, state="disabled"
        )
        self.dl = dl = ttk.Label(self, text="Demangled")
        self.de = de = ttk.Entry(
            self, textvariable=demangled_name, state="readonly", font="TkFixedFont"
        )
        self.db = db = CopyButton(self, demangled_name, state="disabled")

        # https://www.tutorialspoint.com/how-do-i-get-an-event-callback-when-a-tkinter-entry-widget-is-modified
        mangled_name.trace_add("write", self.set_state)

        ml.grid(row=0, column=0, padx=5, pady=10, sticky="w")
        me.grid(row=0, column=1, padx=5, pady=10, sticky="ew")
        mb.grid(row=0, column=2, padx=5, pady=10, sticky="w")
        dl.grid(row=1, column=0, padx=5, pady=(0, 10), sticky="w")
        de.grid(row=1, column=1, padx=5, pady=(0, 10), sticky="ew")
        db.grid(row=1, column=2, padx=5, pady=(0, 10), sticky="w")

        self.place_window_center()
        self.deiconify()
        self.focus_set()
        log.debug("Initialised")

    def set_state(self, *_):
        """Toggles the state of the **Demangle** button based on **Name**."""
        if self.me.get():
            self.mb.configure(state="normal")
        else:
            self.mb.configure(state="disabled")

    def demangle(self):
        """Tries to demangle **Name** and update the **Demangled** entry."""
        mangled = self.mangled_name.get()
        try:
            d = demangle(mangled)
        except DemangleError as exc:
            log.exception(exc)
            Messagebox.show_error(
                f"Failed to demangle '{mangled}'", "Demangling Failed"
            )
        else:
            self.demangled_name.set(d)
            self.db.configure(state="normal")



#!/usr/bin/env python3

"""
dycall.types
~~~~~~~~~~~~

Most analogous to the "Model" part of the MVC design pattern.
"""

from __future__ import annotations

import abc
import dataclasses
import enum
import typing
from ctypes import (
    c_bool,
    c_char,
    c_char_p,
    c_double,
    c_float,
    c_int8,
    c_int16,
    c_int32,
    c_int64,
    c_uint8,
    c_uint16,
    c_uint32,
    c_uint64,
    c_void_p,
    c_wchar,
    c_wchar_p,
)
from typing import Any, Union

try:
    from typing import Final  # type: ignore
except ImportError:
    # pylint: disable=ungrouped-imports
    from typing_extensions import Final  # type: ignore

from dycall.util import DemangleError, demangle

_CType = Union[
    c_bool,
    c_char,
    c_char_p,
    c_double,
    c_float,
    c_int8,
    c_int16,
    c_int32,
    c_int64,
    c_uint8,
    c_uint16,
    c_uint32,
    c_uint64,
    None,
    c_void_p,
    c_wchar,
    c_wchar_p,
]

_StrToCtype = {
    "bool": c_bool,
    "char": c_char,
    "char*": c_char_p,
    "double": c_double,
    "float": c_float,
    "int8_t": c_int8,
    "int16_t": c_int16,
    "int32_t": c_int32,
    "int64_t": c_int64,
    "uint8_t": c_uint8,
    "uint16_t": c_uint16,
    "uint32_t": c_uint32,
    "uint64_t": c_uint64,
    "void": c_void_p,
    "void*": c_void_p,
    "wchar_t": c_wchar,
    "wchar_t*": c_wchar_p,
}


class ParameterType(enum.Enum):
    """Member names mimic format specifiers from `struct` module."""

    bool_ = "bool"
    """C99/C++ bool type. Note: Use `u32` for Win32 BOOL type."""

    c = "char"
    """A single ASCII character."""

    pc = "char*"
    """Null terminated ASCII string."""

    f = "float"
    """32-bit floating point integer."""

    d = "double"
    """64-bit floating point integer. Analogous to `double`"""

    b = "int8_t"
    """8-bit signed integer. Analogous to: `signed char`, `int8_t` and `SBYTE`."""

    h = "int16_t"
    """16-bit signed integer. Analogous to `signed short` and `int16_t`."""

    i = "int32_t"
    """32-bit signed integer. Analogous to `int32_t`, `DWORD` and `int`."""

    q = "int64_t"
    """64-bit signed integer. Analogous to `int64_t`, `__int64` and `long long`."""

    B = "uint8_t"
    """8-bit unsigned integer. Analogous to `uint8_t`, `BYTE` and `char`."""

    H = "uint16_t"
    """16-bit unsigned integer. Analogous to `uint16_t`, `WORD` and `wchar_t`."""

    I = "uint32_t"
    """32-bit unsigned integer. Analogous to `uint32_t`, `DWORD` and `unsigned int`."""

    Q = "uint64_t"
    """64-bit unsigned integer. Analogous to `uint64_t` and `unsigned long long`."""

    v = "void"
    """Void return type."""

    pv = "void*"
    """Any other custom type or a type not in `ParameterType`."""

    w = "wchar_t"
    """A single UTF-16 character."""

    pw = "wchar_t*"
    """Null terminated UTF-16 string."""

    @property
    def ctype(self):
        """Returns equivalent ctypes data type."""
        return _StrToCtype[self.value]


PARAMETER_TYPES: Final = tuple(t.value for t in ParameterType)


class CallConvention(enum.Enum):
    """Calling convention to use when calling a function.

    Since, Linux libs don't use stdcall, this is required only for Windows.
    """

    Cdecl = "cdecl"
    """Commonly used on all platforms. Most probably you will need this."""

    StdCall = "stdcall"
    """More popular on Windows, all system libraries on Windows use this."""


CALL_CONVENTIONS: Final = tuple(t.value for t in CallConvention)


class Marshaller:
    """Common converter methods for Tkinter <-> Python <-> ctypes interop."""

    @staticmethod
    def ctype2str(p: _CType) -> str:
        """Ctypes -> Tkinter."""
        v = "NULL"
        if p is not None:
            val = p.value
            if isinstance(p, c_bool):
                v = "True" if val else "False"
            elif isinstance(p, (c_char, c_char_p)):
                if typing.TYPE_CHECKING:
                    assert isinstance(val, bytes)  # nosec
                v = val.decode("utf-8", errors="replace")
            else:
                v = str(val)
        return v

    @staticmethod
    def pytype2str(p: Any) -> str:
        """Python -> Tkinter."""
        if p:
            if isinstance(p, bytes):
                return p.decode("utf-8", errors="replace")
            return str(p)
        return "NULL"

    @typing.no_type_check
    @staticmethod
    def str2ctype(t: _CType, val: str) -> _CType:
        """Tkinter -> ctypes."""
        # pylint: disable=no-else-return
        if t in (
            c_int8,
            c_int16,
            c_int32,
            c_int64,
            c_uint8,
            c_uint16,
            c_uint32,
            c_uint64,
        ):
            return t(int(val))
        elif t in (c_double, c_float):
            return t(float(val))
        elif t in (c_char, c_char_p):
            return t(val.encode("utf-8"))  # or ascii?
        elif t in (c_wchar, c_wchar_p):
            return t(val)
        return None


@dataclasses.dataclass
class RunResult:
    """Returned by `Runner` back to UI. This is useful especially in **OUT Mode**."""

    ret: _CType
    args: tuple[_CType] = dataclasses.field(default_factory=tuple)  # type: ignore

    @property
    def values(self) -> list[Any]:
        """Ctypes -> Tkinter representation for all arguments."""
        values = []
        for arg in self.args:
            v = Marshaller.ctype2str(arg)
            values.append(v)
        return values


class SortOrder(enum.Enum):
    """Export name sort order in **Exports** combobox."""

    NameAscending = "Name (ascending)"
    NameDescending = "Name (descending)"


@dataclasses.dataclass
class Export(abc.ABC):
    """Constains all the required information about an exported function.

    LIEF doesn't provide a standard platform-independent way to get this
    information. Its better to have a DyCall-specific type; also it is
    used almost everywhere.
    """

    address: int
    """See `lief.Function.address`."""

    name: str
    """See `lief.Function.name`."""


@dataclasses.dataclass
class PEExport(Export):
    """Represents a Windows library export."""

    ordinal: int
    """See `lief.PE.ExportEntry.ordinal`."""

    def __post_init__(self):
        """Calculate a demangled/displayed name."""
        if self.name:
            try:
                self.demangled_name = demangle(self.name)
            except DemangleError as exc:
                self.exc = exc
        else:
            # Ordinal-only exports
            self.demangled_name = f"@{self.ordinal}"


@dataclasses.dataclass
class ELFExport(Export):
    """Represents a Linux or MacOS library export."""

    demangled_name: str
    """See `lief.ELF.Symbol.demangled_name`."""



#!/usr/bin/env python3

"""
dycall.__main__
~~~~~~~~~~~~~~~

Entry point. Command line arguments are parsed and passed over to `App`.
"""

import argparse
import logging
import platform

import desktop_app
import lief

from dycall.app import App
from dycall.types import CALL_CONVENTIONS, PARAMETER_TYPES
from dycall.util import LCIDS

desktop_app.set_process_appid("dycall")
is_windows = platform.system() == "Windows"


# https://stackoverflow.com/a/18700817
def positive_int(s: str) -> int:
    """Positive integer validator for `argparse.ArgumentParser`."""
    i = int(s)
    if i < 0:
        raise argparse.ArgumentTypeError("A positive number is required")
    return i


def main():
    """Arguments are parsed here and passed as keyword arguments."""
    # * Don't use default values for string arguments
    ap = argparse.ArgumentParser(
        prog="DyCall",
        description="Run exported symbols from native libraries",
    )
    ap.add_argument("--log", help="Display logs", action="store_true")
    ap.add_argument("--lib", help="Path/name of library to load on startup.")
    ap.add_argument("--exp", help="Name of export to select from lib.")
    if is_windows:
        ap.add_argument(
            "--conv", help="Calling convention to use.", choices=CALL_CONVENTIONS
        )
    ap.add_argument(
        "--ret", help="Return type of the function.", choices=PARAMETER_TYPES
    )
    ap.add_argument(
        "--rows",
        default=0,
        help="Number of empty rows to create in arguments table",
        type=positive_int,
    )
    ap.add_argument("--lang", help="The language used by the interface", choices=LCIDS)
    ap.add_argument("--out-mode", help="Use 'out' mode", action="store_true")
    ap.add_argument(
        "--hide-errno", help="Hides errno from the status bar", action="store_true"
    )
    ap.add_argument(
        "--no-images",
        action="store_true",
        help="Run without showing any images in the UI.",
    )
    if is_windows:
        ap.add_argument(
            "--hide-gle",
            help="Hides GetLastError from the status bar",
            action="store_true",
        )

    args = ap.parse_args()
    if args.log:
        logging.basicConfig(level=logging.DEBUG)
    else:
        lief.logging.disable()

    launch_args = vars(args)
    _ = launch_args.pop("log", None)  # Logging is handled hee itself
    App(**launch_args).mainloop()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
dycall.top_menu
~~~~~~~~~~~~~~~

Contains `TopMenu`.
"""

import collections
import logging
import platform

import ttkbootstrap as tk
from ttkbootstrap.localization import MessageCatalog as MsgCat

from dycall.about import AboutWindow
from dycall.demangler import DemanglerWindow
from dycall.types import SortOrder
from dycall.util import Lang2LCID, LCID2Lang, get_img

log = logging.getLogger(__name__)


class TopMenu(tk.Menu):
    """DyCall's top menu.

    Hierarchy:
    - File
        - Open Recents
    - Options
        - Language
        - Theme
        - OUT Mode
        - Show GetLastError (Windows only)
        - Show errno
    - View
        - Sort Exports By
            - Name (ascending)
            - Name (descending)
    - Tools
        - Demangler
    - Help
        - About

    `F11` for toggling fullscreen.
    """

    def __init__(
        self,
        root: tk.Window,
        outmode: tk.BooleanVar,
        locale: tk.StringVar,
        sort_order: tk.StringVar,
        show_get_last_error: tk.BooleanVar,
        show_errno: tk.BooleanVar,
        about_opened: tk.BooleanVar,
        theme: tk.StringVar,
        recents: collections.deque,
    ):
        super().__init__()
        self.__root = root
        self.__locale = locale
        self.__about_opened = about_opened
        self.__recents = recents
        self.__lang = tk.StringVar(value=LCID2Lang[locale.get()])

        # File
        self.fo = fo = tk.Menu()
        self.add_cascade(menu=fo, label="File", underline=0)

        # File -> Open Recent
        self.fop = fop = tk.Menu()
        self.__clock_png = get_img("clock.png")
        fo.add_cascade(
            menu=fop,
            label="Open Recent",
            underline=5,
            image=self.__clock_png,
            compound="left",
        )
        self.bind_all("<<UpdateRecents>>", lambda *_: self.update_recents(True))
        self.update_recents()

        # Options
        self.mo = mo = tk.Menu()
        self.add_cascade(menu=mo, label=MsgCat.translate("Options"), underline=0)

        # Options -> Language
        self.mol = mol = tk.Menu(mo)
        self.__translate_png = get_img("translate.png")
        for lang in LCID2Lang.values():
            mol.add_radiobutton(
                label=lang,
                variable=self.__lang,
                command=self.change_lang,
            )
        mo.add_cascade(
            menu=mol,
            label=MsgCat.translate("Language"),
            image=self.__translate_png,
            compound="left",
        )

        # Options -> Theme
        self.mot = mot = tk.Menu(mo)
        self.__theme_png = get_img("theme.png")
        for label in ("System", "Light", "Dark"):
            mot.add_radiobutton(
                label=label,
                variable=theme,
                command=lambda: root.event_generate("<<ThemeChanged2>>"),
            )
        mo.add_cascade(
            menu=mot,
            label=MsgCat.translate("Theme"),
            image=self.__theme_png,
            compound="left",
        )

        # Options -> OUT mode
        mo.add_checkbutton(label=MsgCat.translate("OUT Mode"), variable=outmode)

        # Options -> Show GetLastError
        if platform.system() == "Windows":
            mo.add_checkbutton(
                label=MsgCat.translate("Show GetLastError"),
                variable=show_get_last_error,
                command=lambda: root.event_generate(
                    "<<ToggleGetLastError>>", state=int(show_get_last_error.get())
                ),
            )

        # Options -> Show errno
        mo.add_checkbutton(
            label=MsgCat.translate("Show errno"),
            variable=show_errno,
            command=lambda: root.event_generate(
                "<<ToggleErrno>>", state=int(show_errno.get())
            ),
        )

        # View
        self.vt = vt = tk.Menu()
        self.add_cascade(menu=vt, label=MsgCat.translate("View"), underline=0)

        # View -> Sort Exports By
        self.vse = vse = tk.Menu()
        self.__sort_png = get_img("sort.png")
        self.__sort_name_asc_png = get_img("sort_name_asc.png")
        self.__sort_name_desc_png = get_img("sort_name_desc.png")
        sorter_imgs = (
            self.__sort_name_asc_png,
            self.__sort_name_desc_png,
        )
        for sorter, img in zip(SortOrder, sorter_imgs):
            vse.add_radiobutton(
                label=MsgCat.translate(sorter.value),
                variable=sort_order,
                command=lambda: root.event_generate("<<SortExports>>"),
                image=img,
                compound="left",
            )
        vt.add_cascade(
            menu=vse,
            label=MsgCat.translate("Sort Exports By"),
            image=self.__sort_png,
            compound="left",
        )

        # Tools
        self.mt = mt = tk.Menu()
        self.add_cascade(menu=mt, label=MsgCat.translate("Tools"), underline=0)

        # Tools -> Demangler
        mt.add_command(label="Demangler", command=lambda *_: DemanglerWindow(root))

        # Help
        self.mh = mh = tk.Menu()
        self.add_cascade(menu=mh, label=MsgCat.translate("Help"), underline=0)

        # Help -> About
        self.__info_png = get_img("info.png")
        mh.add_command(
            accelerator="F1",
            command=lambda *_: self.open_about(),
            compound="left",
            image=self.__info_png,
            label=MsgCat.translate("About"),
        )
        self.bind_all("<F1>", lambda *_: self.open_about())

    def change_lang(self):
        """Instructs Tk to change the underlying locale.

        Generates:
            <<LanguageChanged>>: The UI is reinitialised by `dycall.app.App`.
        """
        log.debug("Changing language")
        lc = self.__locale
        lc.set(Lang2LCID[self.__lang.get()])
        MsgCat.locale(lc.get())
        self.__root.event_generate("<<LanguageChanged>>")
        log.info("Changed locale to '%s'", MsgCat.locale())

    def update_recents(self, redraw: bool = False):
        """Clears (optionally) and repopulates **File** -> **Open Recents**.

        Args:
            redraw (bool, optional): Whether to clear the existing list of
                recents. Defaults to False.

        TODO: Something is wrong in this or elsewhere.
        """
        if redraw:
            self.fop.delete(0, 9)
        for path in self.__recents:
            # pylint: disable=cell-var-from-loop
            self.fop.add_command(
                label=path, command=lambda *_: self.__root.picker.load(path=path)
            )

    def open_about(self):
        """Opens the **About** window in a singleton pattern."""
        if not self.__about_opened.get():
            self.__about_opened.set(True)
            AboutWindow(self.__root, self.__about_opened)



#!/usr/bin/env python3

"""
dycall.picker
~~~~~~~~~~~~~

Contain `PickerFrame`.
"""

from __future__ import annotations

import collections
import ctypes.util
import logging
import platform
from tkinter import filedialog

import lief
import ttkbootstrap as tk
from ttkbootstrap import ttk
from ttkbootstrap.dialogs import Messagebox
from ttkbootstrap.localization import MessageCatalog as MsgCat

from dycall.types import ELFExport, Export, PEExport

log = logging.getLogger(__name__)


class PickerFrame(ttk.Labelframe):
    """Implements the library picker.

    Use command line option `--lib` to auto-load a library on app launch.

    Features:
    - Validation
    - Supports short names (thanks to system search order)
    - Remembers recently opened files.
    """

    def __init__(
        self,
        root: tk.Window,
        lib_path: tk.StringVar,
        selected_export: tk.StringVar,
        output: tk.StringVar,
        status: tk.StringVar,
        is_loaded: tk.BooleanVar,
        is_native: tk.BooleanVar,
        exports: list[Export],
        recents: collections.deque,
    ):
        log.debug("Initialising")

        super().__init__(text="Library")
        self.__root = root
        self.__lib_path = lib_path
        self.__selected_export = selected_export
        self.__output = output
        self.__status = status
        self.__is_loaded = is_loaded
        self.__is_native = is_native
        self.__exports = exports
        self.__recents = recents
        self.__os_name = platform.system()

        # Library path entry
        self.le = le = ttk.Entry(
            self,
            textvariable=self.__lib_path,
            validate="focusout",
            validatecommand=(self.register(self.validate), "%P"),
        )
        le.bind("<Return>", self.on_return_key)
        le.pack(fill="x", expand=True, side="left", padx=5, pady=5)

        # Button to invoke file picker
        self.fb = fb = ttk.Button(
            self, text=MsgCat.translate("Browse"), command=self.browse
        )
        fb.pack(side="right", padx=(0, 5), pady=5)

        if lib_path.get():
            self.load()

        log.debug("Initialised")

    def on_return_key(self, *_) -> None:
        """Callback for handling Enter key pressed in **Library** entry."""
        # Without this, a false Export not found occurs
        self.__selected_export.set("")
        self.load()
        self.le.icursor("end")

    def validate(self, s: str) -> bool:
        """Picker entry validation logic.

        Toggles the state of **Exports** and **Function** frame depending upon
        the result of the validation.
        """
        if s:
            ret = ctypes.util.find_library(s)
            if ret:
                if ret == self.__lib_path.get():
                    # Enable
                    self.__root.event_generate("<<ToggleExportsFrame>>", state=1)
                else:
                    self.load(path=ret)
            else:
                # Disable
                self.__root.event_generate("<<ToggleExportsFrame>>", state=0)
                self.__root.event_generate("<<ToggleFunctionFrame>>", state=0)
            return bool(ret)
        return True

    def browse(self) -> None:
        """Opens a file picker to select a library."""
        file = filedialog.askopenfilename(
            title="Select a binary to load",
            filetypes=[
                ("All files", "*.*"),
                ("PE DLL", "*.dll"),
                ("ELF shared object", "*.so"),
                ("MachO dynamic library", "*.dylib"),
            ],
        )
        if file:
            # Without this, a false Export not found occurs
            self.__selected_export.set("")
            self.load(True, file)

    def load(self, dont_search: bool = False, path: str = None) -> None:
        """Implements the library load logic.

        1. Finds absolute path if `dont_search` is `False`.
        2. Parses the library with LIEF.
        3. Checks if it is a native library.
        4. Triggers the population of **Exports** combobox.
        5. Updates recents.

        Args:
            dont_search (bool, optional): Don't use the system search
                order for finding the library. Defaults to False.
            path (str, optional): Used instead of `self.__lib_path` as the library
                path. Defaults to None. If None, then `self.__lib_path` is used.
        """

        def failure():
            self.__is_loaded.set(False)
            self.__status.set("Load failed")
            Messagebox.show_error(f"Failed to load binary {path}", "Load failed")

        # Find absolute path
        if path is not None:
            self.__lib_path.set(path)
        else:
            path = self.__lib_path.get()
        if not dont_search:
            abspath = ctypes.util.find_library(path)
            if abspath is not None:
                path = abspath
                self.__lib_path.set(path)
        self.__output.set("")

        # * LIEF doesn't raise exceptions
        lib = lief.parse(path)
        if not isinstance(lib, lief.Binary):
            failure()
            return
        self.__is_loaded.set(True)
        self.__status.set("Loaded successfully")
        self.__root.event_generate("<<SetWindowTitle>>")

        os = self.__os_name
        fmt = lib.format
        fmts = lief.EXE_FORMATS
        if (
            (os == "Windows" and fmt == fmts.PE)
            or (os == "Darwin" and fmt == fmts.MACHO)
            or (os == "Linux" and fmt == fmts.ELF)
        ):
            self.__is_native.set(True)
        else:
            Messagebox.show_warning(
                f"{path} is not a native binary. You can view "
                "the exported functions but cannot call them.",
                "Not a native binary",
            )
            self.__is_native.set(False)

        self.__exports.clear()
        if fmt == fmts.PE:
            for exp in lib.get_export().entries:
                self.__exports.append(PEExport(exp.address, exp.name, exp.ordinal))
        elif fmt == fmts.ELF:
            for exp in lib.exported_symbols:
                self.__exports.append(
                    ELFExport(exp.value, exp.name, exp.demangled_name)
                )
        self.__root.event_generate("<<PopulateExports>>")

        # Update recents
        if path not in self.__recents:
            self.__recents.append(path)
        else:
            self.__recents.remove(path)
            self.__recents.appendleft(path)
        self.__root.event_generate("<<UpdateRecents>>")



#!/usr/bin/env python3

"""
dycall.function
~~~~~~~~~~~~~~~

Contains `FunctionFrame`.
"""

from __future__ import annotations

import logging
import platform
import queue
from typing import NamedTuple

import tksheet
import ttkbootstrap as tk
from ttkbootstrap import ttk
from ttkbootstrap.localization import MessageCatalog as MsgCat

from dycall.runner import Runner
from dycall.types import CALL_CONVENTIONS, PARAMETER_TYPES, Marshaller, RunResult
from dycall.util import DARK_THEME

log = logging.getLogger(__name__)


class FunctionFrame(ttk.Frame):
    """

    Command line arguments:
        - `--conv` for **Calling Convention**.
        - `--ret` for **Returns** (return type).
        - `--rows-to-add` for empty rows to add to the **Arguments** table.

    Contains:
        - **Calling Convention** combobox (Windows only)
        - **Returns** combobox
        - **Arguments** table (referred below as tksheet also)
    """

    def __init__(
        self,
        root: tk.Window,
        call_conv: tk.StringVar,
        returns: tk.StringVar,
        lib_path: tk.StringVar,
        export: tk.StringVar,
        output: tk.StringVar,
        status: tk.StringVar,
        is_outmode: tk.BooleanVar,
        is_running: tk.BooleanVar,
        exc_type: tk.StringVar,
        get_last_error: tk.IntVar,
        show_get_last_error: tk.BooleanVar,
        errno: tk.IntVar,
        show_errno: tk.BooleanVar,
        rows_to_add: int,
    ):
        super().__init__()
        self.__root = root
        self.__call_conv = call_conv
        self.__returns = returns
        self.__lib_path = lib_path
        self.__export = export
        self.__output = output
        self.__status = status
        self.__is_outmode = is_outmode
        self.__is_running = is_running
        self.__exc_type = exc_type
        self.__get_last_error = get_last_error
        self.__show_get_last_error = show_get_last_error
        self.__errno = errno
        self.__show_errno = show_errno
        self.__res_q = queue.Queue()  # type: ignore
        self.__exc_q = queue.Queue()  # type: ignore
        self.__args: list[list[str]] = []
        self.__is_windows = platform.system() == "Windows"

        # Call convention
        if self.__is_windows:
            cg = ttk.Labelframe(self, text=MsgCat.translate("Calling Convention"))
            self.cc = cc = ttk.Combobox(
                cg,
                values=CALL_CONVENTIONS,
                textvariable=call_conv,
                state="disabled",
                font=("Courier", 9),
            )
            if not call_conv.get():
                cc.current(0)  # CallConvention.cdecl

        # Return type
        rg = ttk.Labelframe(self, text=MsgCat.translate("Returns"))
        self.rc = rc = ttk.Combobox(
            rg,
            values=PARAMETER_TYPES,
            textvariable=returns,
            state="disabled",
            font=("Courier", 9),
        )
        if not returns.get():
            rc.current(7)  # ParameterType.i (int32_t)

        # Run
        self.rb = rb = ttk.Button(
            self,
            text=f"{MsgCat.translate('Run')}\n(F5)",
            state="disabled",
            command=lambda *_: self.run(),
        )

        # Arguments table
        self.ag = ag = ttk.Labelframe(self, text=MsgCat.translate("Arguments"))
        self.at = at = tksheet.Sheet(
            ag,
            headers=[MsgCat.translate("Type"), MsgCat.translate("Value")],
            empty_horizontal=20,
            row_height=25,
            data=self.__args,
            paste_insert_column_limit=True,
            show_top_left=False,
        )
        if root.style.theme_use() == DARK_THEME:
            at.change_theme("dark blue")
        at.bind(
            "<<ThemeChanged>>",
            lambda _: at.change_theme("dark blue")
            if root.style.theme_use() == DARK_THEME
            else at.change_theme(),
        )
        at.extra_bindings("end_edit_cell", self.table_end_edit_cell)
        at.extra_bindings("end_insert_rows", self.table_end_insert_rows)
        at.extra_bindings("end_paste", self.table_end_paste)
        at.readonly_columns(columns=[0])
        at.default_row_height(height=30)
        at.column_width(1, width=250)
        if rows_to_add > 0:
            for row_index in range(rows_to_add):
                at.insert_row()
                self.table_end_insert_rows(row=row_index)

        if self.__is_windows:
            cc.grid(sticky="ew", padx=5, pady=5)
        rc.grid(sticky="ew", padx=5, pady=5)
        at.grid(sticky="nsew", padx=5, pady=5)

        if self.__is_windows:
            cg.grid(row=0, column=0, sticky="ew", padx=5)
            rg.grid(row=0, column=1, sticky="ew")
            rb.grid(row=0, column=2, padx=5)
        else:
            rg.grid(row=0, column=0, sticky="ew", padx=5)
            rb.grid(row=0, column=1, padx=5)
        ag.grid(row=1, columnspan=3, sticky="nsew", padx=5, pady=5)

        self.rowconfigure(1, weight=1)
        self.columnconfigure(0, weight=1)
        if self.__is_windows:
            self.columnconfigure(1, weight=1)

        if self.__is_windows:
            cg.columnconfigure(0, weight=1)
        rg.columnconfigure(0, weight=1)
        ag.rowconfigure(0, weight=1)
        ag.columnconfigure(0, weight=1)

        self.bind_all(
            "<<ToggleFunctionFrame>>", lambda event: self.set_state(event.state == 1)
        )

    def set_state(self, activate: bool = True):
        """Toggles the state of subwidgets."""
        if activate:
            self.rc.configure(state="readonly")
            if self.__is_windows:
                self.cc.configure(state="readonly")
            self.rb.configure(state="normal")
            self.at.enable_bindings()
            for binding in ("rc_insert_column", "rc_delete_column", "cut", "delete"):
                self.at.disable_bindings(binding)
            self.bind_run_button()
        else:
            for w in (self.rc, self.rb):
                w.configure(state="disabled")
            if self.__is_windows:
                self.cc.configure(state="disabled")
            self.at.disable_bindings()
            self.unbind_run_button()

    def table_end_insert_rows(self, event: NamedTuple = None, row: int = None):
        """Callback for tksheet's `end_insert_rows` binding."""
        if event is not None:
            param1 = event[1]
        else:
            param1 = row
        self.at.create_dropdown(
            r=param1,
            c=0,
            set_value="uint32_t",
            values=PARAMETER_TYPES,
            redraw=True,
            selection_function=self.table_dropdown_change,
        )
        self.at.set_cell_data(r=param1, c=1, value="0")

    def table_end_edit_cell(self, event: NamedTuple):
        """Callback for tksheet's `end_edit_cell` binding."""
        row, col, action, text, *_ = event
        if action != "Escape":
            if col == 1:
                self.table_validate(row, text)
            else:
                self.at.dehighlight_cells(row, 1)

    def table_end_paste(self, event: NamedTuple):
        """Callback for tksheet's `end_paste` binding."""
        _, (x, y), content = event
        if isinstance(x, int):
            row, col = x, y
            if col == 1:
                text = content[0][0]
                self.table_validate(row, text)
        elif isinstance(x, str):
            what, which = x, y
            if what == "column" and which == 1:
                for idx, field in enumerate(content):
                    # Idk why field is a list itself
                    self.table_validate(idx, field[0])

    def table_validate(self, row: int, text: str):
        """Callback for tksheet's `end_insert_rows` binding."""
        t = self.at.get_cell_data(row, 0)
        try:
            # bool & void have readonly cells, no need to validate
            if t in ("float", "double"):
                float(text)
            elif t not in ("char", "char*", "wchar_t", "wchar_t*"):
                int(text)
        # Catch multiple exceptions: https://stackoverflow.com/a/6470452
        except (TypeError, ValueError):
            self.at.highlight_cells(row, 1, bg="red")
        else:
            self.at.dehighlight_cells(row, 1)

    def table_dropdown_change(self, event):
        """Arguments' table dropdown selection callback.

        This method defines the behavior of the **Value** column when new rows
        are added to the table or existing one's **Type** is changed. Editing
        is disabled for all dropdown boxes.

        Behaviors:
            bool: A True and False dropdown is created.
            void: Editing is disabled and value is set to NULL.
            float/double: Value is set to 0.0
            character/string types: Value is cleared.
            integer types: Value is set to 0
        """
        # pylint: disable=no-else-return
        # pylint: disable=bare-except
        row, _, _, type_ = event
        t = self.at

        if type_ == "bool":
            # Check if type is not bool already
            if t.get_dropdown_value(row, 0) != "bool":
                t.create_dropdown(row, 1, values=["True", "False"], redraw=True)
                t.readonly_cells(row, 1)
            return
        else:
            try:
                t.readonly_cells(row, 1, readonly=False)
                t.delete_dropdown(row, 1)
            except KeyError:
                t.set_cell_data(row, 1)

        if type_ == "void":
            t.set_cell_data(row, 1, value="NULL")
            t.readonly_cells(row, 1)
            return
        else:
            t.readonly_cells(row, 1, readonly=False)

        if type_ in ("float", "double"):
            t.set_cell_data(row, 1, value="0.0")
        elif type_ not in ("char", "char*", "void*", "wchar_t", "wchar_t*"):
            t.set_cell_data(row, 1, value="0")

    def process_queue(self):
        """Checks the `RunResult` and exception queues for an element.

        This function schedules itself to run every 100ms in the UI thread
        until either one of the queues has an element, which also means that
        the `dycall.runner.Runner` thread has finished.
        """
        try:
            exc: Exception = self.__exc_q.get_nowait()
        except queue.Empty:
            pass
        else:
            log.exception(exc)
            raise exc

        try:
            result: RunResult = self.__res_q.get_nowait()
        except queue.Empty:
            self.after(100, self.process_queue)
        else:
            self.__root.event_generate("<<OutputSuccess>>")
            self.__status.set("Operation successful")
            ret = Marshaller.pytype2str(result.ret)
            self.__output.set(ret)
            if self.__is_outmode.get():
                self.at.set_column_data(1, result.values, redraw=True)
            self.activate_copy_button()
            self.rb.configure(state="normal")

    def run(self) -> None:
        """Executes the function and updates the UI back with results.

        This function acts as a bridge between the runner and the UI threads.
        Invoked by **Run** button or `F5`.
        """

        def handle_exc(e: Exception, status: str):
            log.exception(e)
            self.__exc_type.set(type(e).__name__)
            # ! Cannot pass an arbitrary string directly even though Tk supports it
            # https://stackoverflow.com/a/21234342
            # https://bugs.python.org/issue3405
            self.__root.event_generate("<<OutputException>>")
            self.__output.set(str(e))
            self.__status.set(status)
            self.activate_copy_button(bootstyle="danger")
            self.rb.configure(state="normal")

        ret_type = self.__returns.get()
        self.__status.set("Running...")
        self.rb.configure(state="disabled")
        self.unbind_run_button()

        try:
            thread = Runner(
                self.__exc_q,
                self.__res_q,
                self.__args,
                self.__call_conv.get(),
                ret_type,
                self.__lib_path.get(),
                self.__export.get(),
                self.__get_last_error,
                self.__show_get_last_error.get(),
                self.__errno,
                self.__show_errno.get(),
            )
        except Exception as e:  # pylint: disable=broad-except
            handle_exc(e, "Invalid argument(s)")
            self.bind_run_button()
            return
        self.__is_running.set(True)
        thread.start()

        try:
            self.process_queue()
        except Exception as e:  # pylint: disable=broad-except
            handle_exc(e, "An error occured")
        self.__is_running.set(False)
        self.bind_run_button()

    # * Helpers
    def activate_copy_button(
        self, state: str = "normal", bootstyle: str = "default"
    ) -> None:
        """Activates and configures the bootstyle of the `CopyButton`."""
        self.__root.output.oc.configure(state=state, bootstyle=bootstyle)

    def bind_run_button(self) -> None:
        """Binds the **Run** button to `F5`."""
        # pylint: disable=attribute-defined-outside-init
        log.debug("Run button binded to F5")
        self.rb.bind_all("<F5>", lambda *_: self.run())

    def unbind_run_button(self) -> None:
        """Unbinds the **Run** button while a function is executing."""
        log.debug("Run button unbinded")
        self.rb.unbind_all("<F5>")



#!/usr/bin/env python3

"""
dycall.status_bar
~~~~~~~~~~~~~~~~~

Contains `StatusBarFrame`.
"""

import logging
import os
import platform

try:
    import win32api
except ImportError:
    pass

import ttkbootstrap as tk
from ttkbootstrap import ttk

from dycall.util import StaticThemedTooltip

log = logging.getLogger(__name__)


class StatusBarFrame(ttk.Frame):
    """Implements a typical desktop application status bar.

    Contains:
        - A useful message which changes according to the current state of the
          application and user events (on the left).
        - Errno and GetLastError (Window only) on the right.

    Command line arguments:
        - `--hide-errno` to hide **errno**.
        - `--hide-gle` to hide **GetLastError** (Windows only).
    """

    def __init__(
        self,
        _: tk.Window,
        status: tk.StringVar,
        get_last_error: tk.IntVar,
        show_get_last_error: tk.BooleanVar,
        errno: tk.IntVar,
        show_errno: tk.BooleanVar,
    ):
        log.debug("Initialising")
        super().__init__()
        self.__show_get_last_error = show_get_last_error
        self.__show_errno = show_errno
        is_windows = platform.system() == "Windows"

        # Status
        self.sb = sb = ttk.Label(self, textvariable=status)
        sb.pack(side="left")

        # Errcodes
        self.eg = eg = ttk.Frame(self)

        if is_windows:
            # GetLastError
            gf = ttk.Frame(eg)
            ttk.Label(gf, text="GetLastError: ").pack(side="left")
            self.gb = gb = ttk.Label(
                gf, textvariable=get_last_error, font="TkFixedFont"
            )
            gf.bind(
                "<Enter>",
                lambda *_: StaticThemedTooltip(
                    gf, lambda *_: win32api.FormatMessageW(get_last_error.get())
                ),
            )
            self.bind_all(
                "<<ToggleGetLastError>>",
                lambda event: gf.grid_forget()
                if event.state == 0
                else gf.grid_configure(row=0, column=0),
            )
            gb.pack(side="right")
            if show_get_last_error.get():
                gf.grid(row=0, column=0)

            # Separator
            self.__sf = sf = ttk.Frame(eg)
            ttk.Separator(sf, orient="vertical").place(relheight=1.0)
            sf.bind_all("<<ToggleErrno>>", lambda *_: self.adjust_separator(), add=True)
            sf.bind_all(
                "<<ToggleGetLastError>>", lambda *_: self.adjust_separator(), add=True
            )
            if show_get_last_error.get() and show_errno.get():
                sf.grid(row=0, column=1, sticky="NS")

        # Errno
        ef = ttk.Frame(eg)
        ttk.Label(ef, text="errno: ").pack(side="left")
        self.eb = eb = ttk.Label(ef, textvariable=errno, font="TkFixedFont")
        ef.bind(
            "<Enter>",
            lambda *_: StaticThemedTooltip(ef, os.strerror(errno.get())),
        )
        self.bind_all(
            "<<ToggleErrno>>",
            lambda event: ef.grid_forget()
            if event.state == 0
            else ef.grid_configure(row=0, column=2),
            add=True,
        )
        eb.pack(side="right")
        if show_errno.get():
            ef.grid(row=0, column=2)

        eg.pack(side="right", fill="y")
        log.debug("Initialised")

    def adjust_separator(self):
        """Reconfigures separator when **errno** or **GetLastError** is hidden.

        Only used on Windows.
        """
        if self.__show_errno.get() and self.__show_get_last_error.get():
            self.__sf.grid_configure(row=0, column=1, sticky="NS")
        else:
            self.__sf.grid_forget()



#!/usr/bin/env python3

"""Tests for images and translations."""

from pathlib import Path

from dycall.util import LCIDS

root = Path(__file__).parent.parent.resolve()


def test_check_translations():
    """Checks if translations exist and are syntactically correct.

    1. Whether all the locale IDs (except English) mentioned in
       `dycall.util.LCIDS` have a translation file.
    2. Checks if the command `::msgcat::mcset` and the correct LCID
       is present every line.
    """
    msgs = root / "dycall/msgs"
    lcids = list(LCIDS)
    lcids.remove("en")
    for lcid in lcids:
        msg = Path(msgs / f"{lcid}.msg")
        assert msg.is_file()
        with open(msg, encoding="utf-8") as fp:
            for line in fp.readlines():
                command, locale = line.strip().split()[:2]
                assert command == "::msgcat::mcset"
                assert locale == lcid


def test_check_images():
    """Checks whether all the images used by DyCall are present in the package."""
    imgs = root / "dycall/img"
    for img in (
        "clock.png",
        "dycall.ico",
        "dycall.png",
        "github.png",
        "info.png",
        "sort.png",
        "sort_name_asc.png",
        "sort_name_desc.png",
        "theme.png",
        "translate.png",
    ):
        assert Path(imgs / img).is_file()



#!/usr/bin/env python3

"""
dycall.util
~~~~~~~~~~~

Contains:
- Demangling: Logic used by `dycall.types.PEExport`, `dycall.types.ELFExport`
  and `dycall.demangler.DemanglerWindow`.
- Constants: TtkBootstrap light and dark theme names.
- Custom widgets: A tooltip and a copy button.
- Helpers: Image path and PhotoImage object getters.
"""

from __future__ import annotations

import ctypes
import logging
import pathlib
import platform
from typing import Callable, Optional, Union

try:
    from typing import Final  # type: ignore
except ImportError:
    # pylint: disable=ungrouped-imports
    from typing_extensions import Final  # type: ignore

try:
    import cxxfilt
except ImportError:
    pass

import tktooltip
import ttkbootstrap as tk
from ttkbootstrap import ttk

log = logging.getLogger(__name__)

# * Demangling

os = platform.system()
BUFSIZE: Final = 1000  # That should probably be enough


class DemangleError(Exception):
    """Raised when demangling fails due to any reason."""


def demangle(exp: str) -> str:
    """On Linux & MacOS, LIEF already provides the demangled name.

    On Windows, the DbgHelp API function `UnDecorateSymbolNameW` is used.
    MSDN: https://docs.microsoft.com/windows/win32/api/dbghelp/nf-dbghelp-undecoratesymbolnamew
    """  # noqa: E501
    if os == "Windows":
        if exp.startswith("?"):
            buf = ctypes.create_unicode_buffer(BUFSIZE)
            try:
                dbghelp = ctypes.windll["dbghelp"]
                hr = dbghelp.UnDecorateSymbolNameW(exp, buf, BUFSIZE, 0)
            except OSError as e:
                raise DemangleError from e
            if hr:
                return buf.value
            raise DemangleError
        return exp
    try:
        return cxxfilt.demangle(exp)
    except cxxfilt.Error as e:
        raise DemangleError from e


# * Constants

LIGHT_THEME: Final = "yeti"
DARK_THEME: Final = "darkly"

# * Custom widgets


class CopyButton(ttk.Button):
    """Button which copies text from a `tk.StringVar` to system clipboard."""

    def __init__(self, parent: tk.Window, copy_from: tk.StringVar, *args, **kwargs):
        self.__copy_var = copy_from
        super().__init__(
            parent,
            text="⧉",
            command=lambda *_: self.copy(),
            style="info-outline",
            *args,
            **kwargs,
        )
        self.bind("<Enter>", lambda *_: StaticThemedTooltip(self, "Copy", delay=0.5))

    def copy(self):
        """Clears the clipboard and appends new text.

        Tkinter's clipboard system works a bit differently.
        """
        self.clipboard_clear()
        self.clipboard_append(self.__copy_var.get())


class StaticThemedTooltip(tktooltip.ToolTip):
    """A non-tracking theme-aware tooltip with a configurable delay."""

    def __init__(
        self,
        widget: tk.tk.Widget,
        msg: Union[str, Callable] = None,
        delay: float = 1,
    ):
        fg = bg = None
        if tk.Style().theme_use() == DARK_THEME:
            fg = "#ffffff"
            bg = "#1c1c1c"
        super().__init__(
            widget=widget,
            msg=msg,
            delay=delay,
            follow=False,
            fg=fg,
            bg=bg,
        )


# * Translations

# ! Translators should add the LCID and native form of the language below
LCID2Lang: Final = {"en": "English", "hi": "हिन्दी", "mr": "मराठी"}

LCIDS: Final = tuple(LCID2Lang.keys())

# Dictionary inversion: https://stackoverflow.com/a/66464410
Lang2LCID: Final = {v: k for k, v in LCID2Lang.items()}

# * Helpers

SHOW_IMAGES = True
"""`get_img_path` and `get_img` return None when this is False.

Set to False when DyCall is run with the `--no-images` switch."""


class _ImageFinder:
    """DyCall image finder.

    Images are searched in the `img/` relative to the folder this file is in.
    I didn't quite like what `pkgutil` or `importlib.resources` had to offer.
    Images should never be loaded/searched without using this class or the
    helper methods `get_img` and `get_img_path`.
    """

    # https://stackoverflow.com/a/3430395
    _dirpath = pathlib.Path(__file__).parent.resolve()
    _imgpath: Final = _dirpath / "img"

    def __init__(self, name: str, **kwargs) -> None:
        self.__name = name
        self.__photo_image_kw = kwargs

    @property
    def path(self) -> str:
        """Returns the absolute path of the image.

        Use this only when `photo_image` is not an option.
        """
        log.debug("Getting path of image %s", self.__name)
        return str(self._imgpath / self.__name)

    @property
    def photo_image(self) -> tk.PhotoImage:
        """Returns the image as a `tk.PhotoImage` object.

        Use this whenever possible.
        """
        log.debug("Getting image object for %s", self.__name)
        with open(self._imgpath / self.__name, "rb") as img:
            return tk.PhotoImage(data=img.read(), **self.__photo_image_kw)


def get_img_path(name: str) -> Optional[str]:
    """Returns the absolute path of an image."""
    if SHOW_IMAGES:
        return _ImageFinder(name).path
    return None


def get_img(name: str, **kwargs) -> Optional[tk.PhotoImage]:
    """Returns an image as a `tk.PhotoImage` object.

    Args:
        name (str): File name of image.
        kwargs: Additional arguments passed directly to `tk.PhotoImage`.
    """
    if SHOW_IMAGES:
        return _ImageFinder(name, **kwargs).photo_image
    return None



#!/usr/bin/env python3

"""Changes a particular color in an image and overwrites it.

Ref: https://stackoverflow.com/a/3850345
"""

from __future__ import annotations

import argparse
import pathlib

from colour import Color
from PIL import Image


def get_rgba(color: str) -> tuple[int, int, int, int]:
    """e.g. black -> (0, 0, 0, 255)."""
    colour = Color(color)
    red = int(colour.get_red() * 255)
    green = int(colour.get_green() * 255)
    blue = int(colour.get_blue() * 255)
    return (red, green, blue, 255)


def main(path: pathlib.Path, src: str, dst: str):  # noqa
    src_rgba = get_rgba(src)
    dst_rgba = get_rgba(dst)
    with Image.open(path) as img:
        pixdata = img.convert("RGBA").load()
        for y in range(img.size[1]):
            for x in range(img.size[0]):
                if pixdata[x, y] == src_rgba:
                    img.putpixel((x, y), dst_rgba)
        img.save(path)


if __name__ == "__main__":
    ap = argparse.ArgumentParser(description=__doc__)
    ap.add_argument("path", type=pathlib.Path)
    ap.add_argument("src", default="black")
    ap.add_argument("dst")
    args = vars(ap.parse_args())
    main(**args)



#!/usr/bin/env python3

"""
dycall.about
~~~~~~~~~~~~

Contains `AboutWindow`.
"""

from __future__ import annotations

import logging
import webbrowser

import requests

try:
    import importlib.metadata as pkg_metadata
except ImportError:
    import importlib_metadata as pkg_metadata  # type: ignore

import ttkbootstrap as tk
from ttkbootstrap import ttk
from ttkbootstrap.dialogs import Messagebox
from ttkbootstrap.localization import MessageCatalog as MsgCat

from dycall.util import get_img

log = logging.getLogger(__name__)


class AboutWindow(tk.Toplevel):
    """DyCall's about page.

    Found under **Help** -> **About** in the top menu.
    Only one instance of this can be open at a time due to its
    singleton nature.

    Shows:
    - Name and installed version of package.
    - Copyright.
    - Hyperlink to GitHub repo.
    - Names and installed versions of dependencies.
    - Button to check for an updated version of DyCall.
    - License notice.
    """

    def __init__(self, _: tk.Window, is_opened: tk.BooleanVar):
        log.debug("Initialising")
        self.__is_opened = is_opened

        super().__init__(
            alpha=0.95,
            title="About",
            topmost=True,
            overrideredirect=True,
        )
        self.withdraw()
        self.resizable(False, False)
        self.bind("<Escape>", lambda *_: self.destroy())

        # Topbar to allow window dragging
        tf = ttk.Frame(self)
        self.cb = cb = ttk.Button(tf, text="🗙", command=lambda *_: self.destroy())
        cb.bind(
            "<Enter>",
            lambda *_: self.after(50, lambda: cb.configure(bootstyle="danger")),
        )
        cb.bind(
            "<Leave>",
            lambda *_: self.after(50, lambda: cb.configure(bootstyle="primary")),
        )
        cb.pack(side="right")
        tf.bind("<ButtonPress-1>", self.start_move)
        tf.bind("<ButtonRelease-1>", self.stop_move)
        tf.bind("<B1-Motion>", self.do_move)
        tf.pack(side="top", fill="x")

        # Project name and version
        mf = ttk.Frame(self)
        ttk.Label(mf, text="DyCall", font=tk.font.Font(size=24)).pack(side="left")
        ttk.Label(
            mf, text=f"v{pkg_metadata.version('dycall')}", font=tk.font.Font(size=10)
        ).pack(side="right", pady=(10, 0))
        mf.pack(pady=10)

        ttk.Label(self, text="(c) demberto 2022").pack(pady=5)

        # Repo link
        # https://stackoverflow.com/a/15216402
        self.__github_png = get_img("github.png")
        gb = ttk.Label(self, cursor="hand2")
        if self.__github_png:
            gb.configure(image=self.__github_png)
        else:
            # Imageless mode
            gb.configure(text="DyCall @ Github", foreground="green")

        # https://stackoverflow.com/a/68306781
        gb.bind(
            "<ButtonRelease-1>",
            lambda *_: webbrowser.open_new_tab("https://github.com/demberto/DyCall"),
        )
        gb.pack(pady=5)

        # Dependencies
        requirements: list[tuple[str, str]] = []
        for requirement in pkg_metadata.requires("dycall"):  # type: ignore
            package = requirement.split(">")[0]
            try:
                requirements.append((package, pkg_metadata.version(package)))
            except pkg_metadata.PackageNotFoundError:
                pass
        tv = ttk.Treeview(
            self,
            columns=("Package", "Version"),
            show="tree",
            height=len(requirements),
            selectmode="none",
        )
        tv.column("#0", width=0, stretch=False)
        tv.column("Package", width=150)
        tv.column("Version", width=50)
        for package, version in requirements:
            tv.insert("", "end", values=(package, version))
        tv.pack(pady=5, fill="x", padx=5)

        # Update checker
        self.__ubt = tk.StringVar(value=MsgCat.translate("Check for updates"))
        self.__ub = ub = ttk.Button(
            self,
            textvariable=self.__ubt,
            command=lambda *_: self.check_for_updates(),
        )
        ub.pack(pady=10)

        # License notice
        ll = ttk.Label(
            self,
            text=MsgCat.translate("DyCall is distributed under the MIT license"),
            font=tk.font.Font(size=9),
        )
        ll.pack(side="bottom", pady=(0, 5))

        width = max(ll.winfo_reqwidth(), 250)
        self.geometry(f"{width}x410")
        self.place_window_center()
        self.deiconify()
        self.focus_set()
        self.grab_set()
        log.debug("Initialised")

    def destroy(self) -> None:
        """Resets singleton lock and exits."""
        self.__is_opened.set(False)
        return super().destroy()

    def start_move(self, event: tk.tk.Event):
        """Window drag start handler."""
        # pylint: disable=attribute-defined-outside-init
        self.x = event.x
        self.y = event.y

    def stop_move(self, _):
        """Window drag stop handler."""
        # pylint: disable=attribute-defined-outside-init
        self.x = None
        self.y = None

    def do_move(self, event: tk.tk.Event):
        """Window dragging action handler."""
        deltax = event.x - self.x
        deltay = event.y - self.y
        x = self.winfo_x() + deltax
        y = self.winfo_y() + deltay
        self.geometry(f"+{x}+{y}")

    def check_for_updates(self):
        """Synchronous update checker using GitHub API."""

        def reset_button():
            self.__ubt.set(MsgCat.translate("Check for updates"))
            self.__ub.configure(bootstyle="primary")

        self.__ubt.set(MsgCat.translate("Checking..."))
        self.__ub.configure(bootstyle="secondary")
        self.update()
        try:
            r = requests.get(
                "https://api.github.com/repos/demberto/DyCall/releases/latest"
            )
        except requests.exceptions.RequestException as exc:
            reset_button()
            log.exception(exc)
            if (
                Messagebox.retrycancel(
                    f"Failed to check for updates! ({type(exc).__name__}). Retry?",
                    "Update check failed",
                    parent=self,
                )
                == "Retry"
            ):
                self.check_for_updates()
        else:
            reset_button()
            release = r.json()
            installed_version = f"v{pkg_metadata.version('dycall')}"
            latest_version = release["tag_name"]
            if installed_version != latest_version:
                if (
                    Messagebox.show_question(
                        "A new version is available. Open the update page?",
                        "Update available",
                        parent=self,
                    )
                    == "Yes"
                ):
                    webbrowser.open_new_tab(
                        "https://github.com/demberto/DyCall/releases/latest"
                    )
            else:
                Messagebox.show_info(
                    "No new updates are available currently.",
                    "No updates available",
                    parent=self,
                )



#!/usr/bin/env python3

"""Tests DyCall UI and its behaviors."""

from __future__ import annotations

import json


def test_default_config(create_app, tmp_path):
    """Tests the configuration file saved on first exit."""
    create_app.destroy()
    settings_file = tmp_path / "settings.json"
    assert settings_file.is_file()
    with open(settings_file, encoding="utf-8") as fp:
        s: dict = json.load(fp)
    assert set(s.keys()) == set(
        (
            "geometry",
            "out_mode",
            "locale",
            "recents",
            "theme",
            "show_errno",
            "show_get_last_error",
        )
    )
